"""Common functions used by the HIL commands"""
import sys
import json
import os.path
import tempfile
import traceback
import subprocess
from datetime import datetime
from os import getcwd
from typing import Optional
import click
from dynaconf import Dynaconf
from embedops_cli.hil.hil_types import get_hil_config_path
from embedops_cli.sse import eo_sse
from embedops_cli import config
from embedops_cli.api.rest import ApiException
from embedops_cli.eo_types import (
    EmbedOpsException,
    NoAvailableHilDevice,
    NoRepoIdException,
    NoCIRunIdException,
    NetworkException,
    UnauthorizedUserException,
)
from embedops_cli.hil.hil_types import (
    NoHILRootPathException,
    HILPackageCreationException,
    NoHILArtifactsPathException,
    HILFilesInvalidException,
)
from embedops_cli.hil.hil_package import create_hil_package
from embedops_cli import embedops_authorization
from embedops_cli.sse.sse_api import SSEApi
from embedops_cli.utilities import (
    echo_error_and_fix,
    get_client as util_get_client,
)
from embedops_cli.utilities import logging_setup
from embedops_cli.hil import hil_package
from embedops_cli.api.models import HilRunCreateProps

HIL_GATEWAY_IMAGE_NAME = "embedops-gateway-image.img"
_logger = logging_setup(__name__)


def get_hil_root_path():
    """get hil_root_path from .embedops/hil/config.yml

    Returns:
        str: the value stored in <repo_root>/.embedops/hil/config.yml:hil_root_path or None
    """
    dot_eo = Dynaconf(
        load_dotenv=False,
        settings_files=[get_hil_config_path()],
        silent_errors=True,
    )
    dot_eo.configure(
        LOADERS_FOR_DYNACONF=[
            "dynaconf.loaders.yaml_loader",
        ]
    )
    return dot_eo.get("hil_root_path")


def get_hil_artifacts_path():
    """get hil_artifacts from .embedops/hil/config.yml

    Returns:
        str: the value stored in <repo_root>/.embedops/hil/config.yml:hil_artifacts or None
    """
    dot_eo = Dynaconf(
        load_dotenv=False,
        settings_files=[get_hil_config_path()],
        silent_errors=True,
    )
    dot_eo.configure(
        LOADERS_FOR_DYNACONF=[
            "dynaconf.loaders.yaml_loader",
        ]
    )
    return dot_eo.get("hil_artifacts")


def get_hil_artifacts_path_from_ci_artifacts_dir(
    hil_artifacts_path: str,
) -> Optional[str]:
    """get hil_artifacts from ./artifacts directory generated by CI

    Returns:
        str?: path of file in artifacts directory that matches filename
            of hil_artifacts_path or None
    """
    artifacts_dir = os.path.join(os.getcwd(), "artifacts")
    file_name = os.path.basename(hil_artifacts_path)

    for root, _, files in os.walk(artifacts_dir):
        if file_name in files:
            absolute_path = os.path.join(root, file_name)
            relative_path = os.path.relpath(absolute_path)
            return relative_path
    return None


def hil_echo_run_banner(ex_pkg_id: str, sdk_path: str) -> None:
    """print the embedops job banner to the console (local and pipeline)

    Args:
        ex_pkg_id (str): repo_id or ci_run_id
        sdk_path (str): path to hil sdk from the repo root.
                        set in .embedops/hil/config.yml:hil_root_path
    """
    git_hash_run = subprocess.run(
        ["git", "rev-parse", "--short", "HEAD"],
        capture_output=True,
        text=True,
        check=False,
    )
    if git_hash_run.returncode != 0:
        git_hash = "Not Available"
    else:
        git_hash = git_hash_run.stdout.strip()

    current_time = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S UTC")
    click.secho("-" * 80, fg="magenta")
    click.secho("Running job      'hil'", err=False, fg="magenta")
    click.secho(f"-> repo id       '{ex_pkg_id}'", err=False, fg="white")
    click.secho(f"-> directory     '{getcwd()}'", err=False, fg="white")
    click.secho(f"-> hil sdk path  '{sdk_path}'", err=False, fg="white")
    click.secho(f"-> git sha       '{git_hash}'", err=False, fg="white")
    click.secho(f"-> timestamp     '{current_time}'", err=False, fg="white")
    click.secho("-" * 80, fg="magenta")
    click.secho("\n")


def hil_run(  # pylint: disable=R0912,R0914,R0915
    local: bool = False, devices=None
) -> int:
    """
    Run hil in either local or CI mode,
    using the current repository as a source.
    """

    try:
        ex_pkg_id = config.settings.run_id
        if not ex_pkg_id and not local:
            raise NoCIRunIdException()
        if local:
            ex_pkg_id = config.get_repo_id()
            if not ex_pkg_id:
                raise NoRepoIdException()

        hil_root_path = get_hil_root_path()
        if not hil_root_path:
            raise NoHILRootPathException()

        hil_artifacts = get_hil_artifacts_path()

        if not local and hil_artifacts:
            hil_artifacts = get_hil_artifacts_path_from_ci_artifacts_dir(hil_artifacts)

        if not hil_artifacts or not os.path.isfile(
            os.path.join(os.path.curdir, hil_artifacts)
        ):
            raise NoHILArtifactsPathException()

        hil_sdk_full_path = os.path.join(os.path.curdir, hil_root_path)
        if not os.path.isdir(hil_sdk_full_path):
            raise NoHILRootPathException()

        _logger.debug(f"using host: {config.settings.host}")
        hil_echo_run_banner(ex_pkg_id, hil_sdk_full_path)

        validate_hil_root(hil_root_path)

        # Compile the package in a temporary folder that is deleted after uploading
        with tempfile.TemporaryDirectory() as tmp_dir:
            hil_ep_file = os.path.join(tmp_dir, "hil_ep.zip")
            # TODO: Manifest data is being left blank intentionally
            if not create_hil_package(
                [os.path.join(os.path.curdir, hil_artifacts)],
                hil_root_path,
                {},
                hil_ep_file,
            ):
                raise HILPackageCreationException()

            # Get the upload URL
            _logger.debug(
                f"[local={local}] getting presigned url for hil execution package upload"
            )
            if local:
                api_client = embedops_authorization.get_user_client()
                upload_url_response = api_client.get_pre_signed_url_for_upload(
                    ex_pkg_id
                )
            else:
                api_client = util_get_client()
                upload_url_response = api_client.get_pre_signed_url_for_upload_0(
                    ex_pkg_id
                )
            upload_url = upload_url_response.url
            try:
                upload_status = hil_package.upload_hil_package(hil_ep_file, upload_url)
                if upload_status != 200:
                    raise NetworkException(
                        upload_status,
                        message="network exception during execution package upload",
                    )
            except ApiException as exc:
                raise EmbedOpsException(
                    fix_message=f"Uploading HIL execution package failed: {exc.body.decode()}"
                ) from exc

        if local:
            # After package is uploaded, call into SSE and print any further events
            sse_api = SSEApi()
            for event in sse_api.sse_hil_run(ex_pkg_id, devices):
                if event.event == eo_sse.SSE_TEXT_EVENT:
                    eo_sse.sse_print_command_text(event)
                elif event.event == eo_sse.SSE_RESULT_EVENT:
                    result_event_obj = json.loads(event.data)
                    return result_event_obj["exitCode"]
                else:
                    _logger.debug(f"got unhandled SSE Event: {event.event}")

            # If the command hasn't returned anything yet, exit here
            return 2

        print("starting hil run...", file=sys.stderr)
        try:
            response = api_client.init_hil_run_from_ci(
                body=HilRunCreateProps(ci_run_id=ex_pkg_id)
            )
            print(f"HIL run started successfully. View progress here: {response.url}")
            return 0
        except ApiException as exc:
            if exc.status == 424:  #
                raise NoAvailableHilDevice(f"{exc.body.decode()}") from exc
            raise exc

    except (
        NoRepoIdException,
        NoHILRootPathException,
        HILPackageCreationException,
        NetworkException,
        UnauthorizedUserException,
        EmbedOpsException,
        NoAvailableHilDevice,
        NoHILArtifactsPathException,
        HILFilesInvalidException,
    ) as exc:
        echo_error_and_fix(exc)
        return 2
    except ApiException as exc:
        echo_error_and_fix(
            EmbedOpsException(
                fix_message=f"Uploading HIL execution package failed: {exc.body.decode()}"
            )
        )
    except Exception:  # pylint: disable=broad-except
        traceback.print_exc()
    return -1


def validate_hil_root(hil_root_path: str) -> bool:
    """Inspect a HIL root path, ensuring that it has a non-zero number of python files,
    and those python files compile"""

    non_sdk_py_files = []

    if not os.path.exists(hil_root_path):
        raise HILFilesInvalidException(
            additional_message="HIL Root path does not exist."
        )

    for path, _, files in os.walk(hil_root_path):
        if "hil_sdk" not in path:
            py_files = [f for f in files if f.endswith(".py")]
            for py_file in py_files:
                non_sdk_py_files.append(os.path.join(path, py_file))

    if len(non_sdk_py_files) == 0:
        raise HILFilesInvalidException(
            additional_message=f"No valid Python files found under {hil_root_path}."
        )

    for py_file in non_sdk_py_files:
        try:
            with open(py_file, "r", encoding="utf-8") as source_file:
                source = source_file.read() + "\n"
                compile(source, py_file, "exec")
        except (SyntaxError, ValueError) as err:
            raise HILFilesInvalidException(additional_message=str(err)) from err
